// 
// Copyright (c) 2010, Benjamin Kaufmann
// 
// This file is part of Clasp. See http://www.cs.uni-potsdam.de/clasp/ 
// 
// Clasp is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// Clasp is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Clasp; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
#ifndef CLASPRE_H_INCLUDED
#define CLASPRE_H_INCLUDED
#include <string>

#if !defined(WITH_CLASPRE)
#	if defined(WITH_CLASPRE_SUPPORT)
#		define WITH_CLASPRE 1
#	else
#		define WITH_CLASPRE 0
#	endif
#endif

namespace ProgramOptions {
class OptionGroup;
}
namespace Clasp { 
class ClaspOptions;
class OutputFormat;
class ClaspreOutput;
class Solver;
namespace Claspre {
// Group "Claspre Options"
struct Options {
	static bool mapFormat(const std::string& s, int& f);
	Options() : features(0), listFeatures(false), hasLimit(false) {}
	enum FeatureFormat {
		features_no        = 0,
		features_verbose   = 1, // like claspre with --features
		features_compact_1 = 2, // like claspre with --claspfolio=1
		features_compact_2 = 3, // like claspre with --claspfolio=2
		features_compact_3 = 4  // like claspre with --claspfolio=3
	};
	void initOptions(ProgramOptions::OptionGroup& root);
	int  features;     // print feature in selected format
	bool listFeatures; // print available features
	bool hasLimit;     // true if "--search-limit" was given
};

void printFeatures();
OutputFormat* createOutput(Solver& s, Options::FeatureFormat f, ClaspOptions& options, int& verbose);

#if !defined(WITH_CLASPRE_SUPPORT)
inline void printFeatures() {}
inline OutputFormat* createOutput(Solver&, Options::FeatureFormat, ClaspOptions&, int&) { return 0; }
inline void Options::initOptions(ProgramOptions::OptionGroup&) {}
#endif
} }
#endif
